import numpy as np
import triangle
import scipy.optimize
from scipy.spatial import Delaunay
import scipy.sparse as sparse
from scipy.sparse import csr_matrix
from scipy.sparse.linalg import factorized
import renderapi
import copy
import os
import datetime
import cv2
from six.moves import urllib
from argschema import ArgSchemaParser
from .schemas import MeshLensCorrectionSchema
from .utils import remove_weighted_matches
from EMaligner import jsongz
import logging

# this is a modification of https://github.com/
# AllenInstitute/render-modules/blob/master/
# rendermodules/mesh_lens_correction/MeshAndSolveTransform.py
# that does not depend on render-modules nor
# on a running render server

logger = logging.getLogger()


class MeshLensCorrectionException(Exception):
    """Exception raised when there is a \
            problem creating a mesh lens correction"""
    pass


def condense_coords(matches):
    # condense point match structure into Nx2
    x = []
    y = []
    for m in matches:
        x += m['matches']['p'][0]
        x += m['matches']['q'][0]
        y += m['matches']['p'][1]
        y += m['matches']['q'][1]
    coords = np.transpose(np.vstack((np.array(x), np.array(y))))
    return coords


def smooth_density(coords, tile_width, tile_height, n):
    # n: area divided into nxn
    min_count = np.Inf
    for i in range(n):
        r = np.arange(
                (i*tile_width/n),
                ((i+1)*tile_width/n))
        for j in range(n):
            c = np.arange(
                    (j*tile_height/n),
                    ((j+1)*tile_height/n))
            ind = np.argwhere(
                    (coords[:, 0] >= r.min()) &
                    (coords[:, 0] <= r.max()) &
                    (coords[:, 1] >= c.min()) &
                    (coords[:, 1] <= c.max())).flatten()
            if ind.size < min_count:
                min_count = ind.size
    new_coords = []
    for i in range(n):
        r = np.arange(
                (i*tile_width/n),
                ((i+1)*tile_width/n))
        for j in range(n):
            c = np.arange(
                    (j*tile_height/n),
                    ((j+1)*tile_height/n))
            ind = np.argwhere(
                    (coords[:, 0] >= r.min()) &
                    (coords[:, 0] <= r.max()) &
                    (coords[:, 1] >= c.min()) &
                    (coords[:, 1] <= c.max())).flatten()
            a = np.arange(ind.size)
            np.random.shuffle(a)
            ind = ind[a[0:min_count]]
            new_coords.append(coords[ind])
    return np.concatenate(new_coords)


def approx_snap_contour(contour, width, height, epsilon=20, snap_dist=5):
    # approximate contour within epsilon pixels,
    # so it isn't too fine in the corner
    # and snap to edges
    approx = cv2.approxPolyDP(contour, epsilon, True)
    for i in range(approx.shape[0]):
        for j in [0, width]:
            if np.abs(approx[i, 0, 0] - j) <= snap_dist:
                approx[i, 0, 0] = j
        for j in [0, height]:
            if np.abs(approx[i, 0, 1] - j) <= snap_dist:
                approx[i, 0, 1] = j
    return approx


def create_PSLG(tile_width, tile_height, maskUrl):
    # define a PSLG for triangle
    # http://dzhelil.info/triangle/definitions.html
    # https://www.cs.cmu.edu/~quake/triangle.defs.html#pslg
    if maskUrl is None:
        vertices = np.array([
                [0, 0],
                [0, tile_height],
                [tile_width, tile_height],
                [tile_width, 0]])
        segments = np.array([
                [0, 1],
                [1, 2],
                [2, 3],
                [3, 0]])
    else:
        mpath = urllib.parse.unquote(
                    urllib.parse.urlparse(maskUrl).path)
        im = cv2.imread(mpath, 0)
        contours, h = cv2.findContours(
                im, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        approx = approx_snap_contour(contours[0], tile_width, tile_height)
        vertices = np.array(approx).squeeze()
        segments = np.zeros((vertices.shape[0], 2))
        segments[:, 0] = np.arange(vertices.shape[0])
        segments[:, 1] = segments[:, 0] + 1
        segments[-1, 1] = 0
    bbox = {}
    bbox['vertices'] = vertices
    bbox['segments'] = segments

    return bbox


def calculate_mesh(a, bbox, target, get_t=False):
    t = triangle.triangulate(bbox, 'pqa%0.1f' % a)
    if get_t:
        # scipy.Delaunay has nice find_simplex method,
        # but, no obvious way to iteratively refine meshes, like triangle
        # numbering is different, so, switch here
        return Delaunay(t['vertices'])
    return target - len(t['vertices'])


def force_vertices_with_npoints(area_par, bbox, coords, npts):
    fac = 1.02
    count = 0
    max_iter = 20
    while True:
        t = calculate_mesh(
                area_par,
                bbox,
                None,
                get_t=True)
        pt_count = count_points_near_vertices(t, coords)
        if pt_count.min() >= npts:
            break
        area_par *= fac
        count += 1
        if np.mod(count, 2) == 0:
            fac += 0.5
        if np.mod(count, max_iter) == 0:
            e = ("did not meet vertex requirement "
                 "after %d iterations" % max_iter)
            raise MeshLensCorrectionException(e)
    return t, area_par


def find_delaunay_with_max_vertices(bbox, nvertex):
    # find bracketing values
    a1 = a2 = 1e6
    t1 = calculate_mesh(a1, bbox, nvertex)
    afac = np.power(10., -np.sign(t1))
    while (
            np.sign(t1) ==
            np.sign(calculate_mesh(a2, bbox, nvertex))
          ):
        a2 *= afac
    val_at_root = -1
    nvtweak = nvertex
    while val_at_root < 0:
        a = scipy.optimize.brentq(
                calculate_mesh,
                a1,
                a2,
                args=(bbox, nvtweak, ))
        val_at_root = calculate_mesh(a, bbox, nvertex)
        a1 = a * 2
        a2 = a * 0.5
        nvtweak -= 1
    mesh = calculate_mesh(a, bbox, None, get_t=True)
    return mesh, a


def compute_barycentrics(coords, mesh):
    # https://en.wikipedia.org/wiki/Barycentric_coordinate_system#Conversion_between_barycentric_and_Cartesian_coordinates
    triangle_indices = mesh.find_simplex(coords)
    vt = np.vstack((
        np.transpose(mesh.points),
        np.ones(mesh.points.shape[0])))
    mt = np.vstack((np.transpose(coords), np.ones(coords.shape[0])))
    bary = np.zeros((3, coords.shape[0]))
    Rinv = []
    for tri in mesh.simplices:
        Rinv.append(np.linalg.inv(vt[:, tri]))
    for i in range(mesh.nsimplex):
        ind = np.argwhere(triangle_indices == i).flatten()
        bary[:, ind] = Rinv[i].dot(mt[:, ind])
    return np.transpose(bary), triangle_indices


def count_points_near_vertices(t, coords):
    flat_tri = t.simplices.flatten()
    flat_ind = np.repeat(np.arange(t.nsimplex), 3)
    v_touches = []
    for i in range(t.npoints):
        v_touches.append(flat_ind[np.argwhere(flat_tri == i)])
    pt_count = np.zeros(t.npoints)
    found = t.find_simplex(coords, bruteforce=True)
    for i in range(t.npoints):
        for j in v_touches[i]:
            pt_count[i] += np.count_nonzero(found == j)
    return pt_count


def create_regularization(ncols, ntiles, defaultL, transL, lensL):
    # regularizations: [affine matrix, translation, lens]
    reg = np.ones(ncols).astype('float64') * defaultL
    reg[0: (ntiles * 3)] *= transL
    reg[(ntiles * 3):] *= lensL
    rmat = sparse.eye(reg.size, dtype='float64', format='csr')
    rmat.data = reg
    return rmat


def create_thinplatespline_tf(
        args, mesh, solution,
        lens_dof_start, common_transform,
        logger,
        compute_affine=False):

    dst = np.zeros_like(mesh.points)
    dst[:, 0] = mesh.points[:, 0] + solution[0][lens_dof_start:]
    dst[:, 1] = mesh.points[:, 1] + solution[1][lens_dof_start:]

    if common_transform is not None:
        # average affine result
        dst = common_transform.tform(dst)

    transform = renderapi.transform.ThinPlateSplineTransform()
    transform.estimate(mesh.points, dst, computeAffine=compute_affine)
    npts0 = transform.srcPts.shape[1]
    transform = transform.adaptive_mesh_estimate(max_iter=1000)
    npts1 = transform.srcPts.shape[1]

    logger.info(
            "adaptive_mesh_estimate reduced control points from %d to %d" %
            (npts0, npts1))

    transform.transformId = (
            datetime.datetime.now().strftime("%Y%m%d%H%M%S"))
    transform.labels = None

    return transform


def new_specs_with_tf(ref_transform, tilespecs, transforms):
    newspecs = []
    for i in range(len(tilespecs)):
        newspecs.append(copy.deepcopy(tilespecs[i]))
        newspecs[-1].tforms.insert(0,
                                   renderapi.transform.ReferenceTransform(
                                    refId=ref_transform.transformId))
        newspecs[-1].tforms[1] = transforms[i]
    return newspecs


def solve(A, weights, reg, x0, b):
    ATW = A.transpose().dot(weights)
    K = ATW.dot(A) + reg
    K_factorized = factorized(K)
    solution = []

    i = 0
    for x in x0:
        Lm = reg.dot(x) + ATW.dot(b[:, i])
        i += 1
        solution.append(K_factorized(Lm))

    errx = A.dot(solution[0]) - b[:, 0]
    erry = A.dot(solution[1]) - b[:, 1]

    return solution, errx, erry


def report_solution(errx, erry, transforms, criteria):
    translation = np.array([tf.translation for tf in transforms])

    jresult = {}
    jresult['x_res_min'] = errx.min()
    jresult['x_res_max'] = errx.max()
    jresult['x_res_mean'] = errx.mean()
    jresult['x_res_std'] = errx.std()
    jresult['y_res_min'] = erry.min()
    jresult['y_res_max'] = erry.max()
    jresult['y_res_mean'] = erry.mean()
    jresult['y_res_std'] = erry.std()

    for k in jresult.keys():
        jresult[k] = np.round(jresult[k], 3)

    message = 'lens solver results [px]'
    for val in ['res']:
        for xy in ['x', 'y']:
            d = '\n%8s' % (xy + '_' + val + ' ')
            v = ''
            for calc in ['min', 'max', 'mean', 'std']:
                d += calc + ','
                k = xy + '_' + val + '_' + calc
                v += '%8.2f' % jresult[k]
            message += d + v

    return translation, jresult, message


def create_x0(nrows, tilespecs):
    ntiles = len(tilespecs)
    x0 = []
    x0.append(np.zeros(nrows).astype('float64'))
    x0.append(np.zeros(nrows).astype('float64'))
    x0[0][0:ntiles] = np.zeros(ntiles)
    x0[1][0:ntiles] = np.zeros(ntiles)
    for i in range(ntiles):
        x0[0][i] = tilespecs[i].tforms[0].B0
        x0[1][i] = tilespecs[i].tforms[0].B1
    return x0


def create_A(matches, tilespecs, mesh):
    # let's assume translation halfsize
    dof_per_tile = 1
    dof_per_vertex = 1
    vertex_per_patch = 3
    nnz_per_row = 2*(dof_per_tile + vertex_per_patch * dof_per_vertex)
    nrows = sum([len(m['matches']['p'][0]) for m in matches])
    nd = nnz_per_row*nrows
    lens_dof_start = dof_per_tile*len(tilespecs)

    data = np.zeros(nd).astype('float64')
    b = np.zeros((nrows, 2)).astype('float64')
    indices = np.zeros(nd).astype('int64')
    indptr = np.zeros(nrows+1).astype('int64')
    indptr[1:] = np.arange(1, nrows+1)*nnz_per_row
    weights = np.ones(nrows).astype('float64')

    unique_ids = np.array(
            [t.tileId for t in tilespecs])

    # nothing fancy here, row-by-row
    offset = 0
    rows = 0
    for mi in range(len(matches)):
        m = matches[mi]
        pindex = np.argwhere(unique_ids == m['pId'])
        qindex = np.argwhere(unique_ids == m['qId'])

        npoint_pairs = len(m['matches']['q'][0])
        # get barycentric coordinates ready
        pcoords = np.transpose(
                np.vstack(
                    (m['matches']['p'][0],
                     m['matches']['p'][1])
                    )).astype('float64')
        qcoords = np.transpose(
                np.vstack(
                    (m['matches']['q'][0],
                     m['matches']['q'][1])
                    )).astype('float64')

        b[rows: (rows + pcoords.shape[0])] = qcoords - pcoords
        rows += pcoords.shape[0]
        pbary = compute_barycentrics(pcoords, mesh)
        qbary = compute_barycentrics(qcoords, mesh)

        mstep = np.arange(npoint_pairs) * nnz_per_row + offset

        data[mstep + 0] = 1.0
        data[mstep + 1] = -1.0
        data[mstep + 2] = pbary[0][:, 0]
        data[mstep + 3] = pbary[0][:, 1]
        data[mstep + 4] = pbary[0][:, 2]
        data[mstep + 5] = -qbary[0][:, 0]
        data[mstep + 6] = -qbary[0][:, 1]
        data[mstep + 7] = -qbary[0][:, 2]

        indices[mstep + 0] = pindex
        indices[mstep + 1] = qindex
        indices[mstep + 2] = (lens_dof_start +
                              mesh.simplices[pbary[1][:]][:, 0])
        indices[mstep + 3] = (lens_dof_start +
                              mesh.simplices[pbary[1][:]][:, 1])
        indices[mstep + 4] = (lens_dof_start +
                              mesh.simplices[pbary[1][:]][:, 2])
        indices[mstep + 5] = (lens_dof_start +
                              mesh.simplices[qbary[1][:]][:, 0])
        indices[mstep + 6] = (lens_dof_start +
                              mesh.simplices[qbary[1][:]][:, 1])
        indices[mstep + 7] = (lens_dof_start +
                              mesh.simplices[qbary[1][:]][:, 2])

        offset += npoint_pairs*nnz_per_row

    A = csr_matrix((data, indices, indptr), dtype='float64')

    wts = sparse.eye(weights.size, format='csr', dtype='float64')
    wts.data = weights
    return A, wts, b, lens_dof_start


def create_transforms(ntiles, solution, get_common=False):
    rtransforms = []
    for i in range(ntiles):
        rtransforms.append(renderapi.transform.AffineModel(
                           B0=solution[0][i],
                           B1=solution[1][i]))

    if get_common:
        transforms = []
        # average without translations
        common = np.array([t.M for t in rtransforms]).mean(0)
        common[0, 2] = 0.0
        common[1, 2] = 0.0
        for r in rtransforms:
            transforms.append(renderapi.transform.AffineModel())
            transforms[-1].M = r.M.dot(np.linalg.inv(common))
        ctform = renderapi.transform.AffineModel()
        ctform.M = common
    else:
        ctform = None
        transforms = rtransforms

    return ctform, transforms


def estimate_stage_affine(t0, t1):
    src = np.array([t.tforms[0].translation for t in t0])
    dst = np.array([t.tforms[1].translation for t in t1])
    aff = renderapi.transform.AffineModel()
    aff.estimate(src, dst)
    return aff


class MeshAndSolveTransform(ArgSchemaParser):
    default_schema = MeshLensCorrectionSchema

    def run(self):
        jspecs = jsongz.load(self.args['tilespec_file'])
        self.tilespecs = np.array([
                renderapi.tilespec.TileSpec(json=j) for j in jspecs])

        self.matches = jsongz.load(self.args['match_file'])

        remove_weighted_matches(self.matches, weight=0.0)

        self.tile_width = self.tilespecs[0].width
        self.tile_height = self.tilespecs[0].height
        maskUrl = self.tilespecs[0].ip[0].maskUrl

        # condense coordinates
        self.coords = condense_coords(self.matches)
        nc0 = self.coords.shape[0]
        self.coords = smooth_density(
            self.coords,
            self.tile_width,
            self.tile_height,
            10)
        nc1 = self.coords.shape[0]
        self.logger.info(
                "\n  smoothing point density reduced points from %d to %d" %
                (nc0, nc1))
        if self.coords.shape[0] == 0:
            raise MeshLensCorrectionException(
                    "no point matches left after smoothing density, \
                     probably some sparse areas of matching")

        # create PSLG
        self.bbox = create_PSLG(
                self.tile_width,
                self.tile_height,
                maskUrl)

        # find delaunay with max vertices
        self.mesh, self.area_triangle_par = \
            find_delaunay_with_max_vertices(
                self.bbox,
                self.args['nvertex'])

        # and enforce neighboring matches to vertices
        self.mesh, self.area_triangle_par = \
            force_vertices_with_npoints(
                self.area_triangle_par,
                self.bbox,
                self.coords,
                3)

        nend = self.mesh.points.shape[0]

        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.info(
                "\n  aimed for %d mesh points, got %d" %
                (self.args['nvertex'], nend))

        if self.mesh.points.shape[0] < 0.5*self.args['nvertex']:
            raise MeshLensCorrectionException(
                    "mesh coarser than intended")

        # prepare the linear algebra and solve
        self.A, self.weights, self.b, self.lens_dof_start = \
            create_A(
                self.matches,
                self.tilespecs,
                self.mesh)
        self.x0 = create_x0(
                self.A.shape[1],
                self.tilespecs)

        self.reg = create_regularization(
                self.A.shape[1],
                len(self.tilespecs),
                self.args['regularization']['default_lambda'],
                self.args['regularization']['translation_factor'],
                self.args['regularization']['lens_lambda'])
        self.solution, self.errx, self.erry = solve(
                self.A,
                self.weights,
                self.reg,
                self.x0,
                self.b)

        self.common_transform, self.transforms = create_transforms(
                len(self.tilespecs), self.solution)

        tf_trans, jresult, self.solve_message = report_solution(
                self.errx,
                self.erry,
                self.transforms,
                self.args['good_solve'])

        self.logger.info(self.solve_message)

        # check quality of solution
        if not all([
                    self.errx.mean() <
                    self.args['good_solve']['error_mean'],

                    self.erry.mean() <
                    self.args['good_solve']['error_mean'],

                    self.errx.std() <
                    self.args['good_solve']['error_std'],

                    self.erry.std() <
                    self.args['good_solve']['error_std']]):

            raise MeshLensCorrectionException(
                    "Solve not good: %s" % self.solve_message)

        self.logger.debug(self.solve_message)

        self.new_ref_transform = create_thinplatespline_tf(
                self.args,
                self.mesh,
                self.solution,
                self.lens_dof_start,
                self.common_transform,
                self.logger)

        bbox = self.tilespecs[0].bbox_transformed(tf_limit=0)
        tbbox = self.new_ref_transform.tform(bbox)
        bstr = 'new transform corners:\n'
        for i in range(bbox.shape[0]-1):
            bstr += "  (%0.1f, %0.1f) -> (%0.1f, %0.1f)\n" % (
                    bbox[i, 0], bbox[i, 1],
                    tbbox[i, 0], tbbox[i, 1])
        self.logger.info(bstr)

        self.tfpath = os.path.join(
                self.args['output_dir'], self.args['outfile'])
        jsongz.dump(
                self.new_ref_transform.to_dict(), self.tfpath, compress=False)

        new_tilespecs = new_specs_with_tf(
            self.new_ref_transform,
            self.tilespecs,
            self.transforms)

        stage_affine = estimate_stage_affine(self.tilespecs, new_tilespecs)
        sastr = "affine estimate of tile translations:\n"
        sastr += "  scale: {}\n".format(stage_affine.scale)
        sastr += "  translation: {}\n".format(stage_affine.translation)
        sastr += "  shear: {}\n".format(stage_affine.shear)
        sastr += "  rotation: {}\n".format(np.degrees(stage_affine.rotation))
        self.logger.info(sastr)

        new_path = os.path.join(
                self.args['output_dir'],
                'resolvedtiles.json')

        resolved = renderapi.resolvedtiles.ResolvedTiles(
                tilespecs=new_tilespecs,
                transformList=[self.new_ref_transform])

        new_path = jsongz.dump(
                resolved.to_dict(),
                new_path,
                compress=self.args['compress_output'])

        self.args['output_json'] = os.path.join(
                self.args['output_dir'],
                'output.json')

        jresult['resolved_tiles'] = os.path.abspath(new_path)

        self.output(jresult, indent=2)

        self.logger.info(" wrote solved tilespecs:\n  %s" %
                         self.args['output_json'])

        return



from argschema.schemas import DefaultSchema
from argschema import ArgSchema
from argschema.fields import (
        InputDir, InputFile, Float,
        Int, OutputFile, Str, Boolean)

# duplicates so that GenerateEMTileSpecsModule looks identical to
# render-modules version


class RenderClientParameters(DefaultSchema):
    host = Str(
        required=True, description='render host')
    port = Int(
        required=True, description='render post integer')
    owner = Str(
        required=True, description='render default owner')
    project = Str(
        required=True, description='render default project')
    client_scripts = Str(
        required=True, description='path to render client scripts')
    memGB = Str(
        required=False,
        default='5G',
        description='string describing java heap memory (default 5G)')
    validate_client = Boolean(
        required=False,
        default=False,
        description="will avoid problems on windows if we use use_rest")


class RenderParameters(argschema.ArgSchema):
    render = argschema.fields.Nested(
        RenderClientParameters,
        required=True,
        description="parameters to connect to render server")


class OverridableParameterSchema(argschema.schemas.DefaultSchema):
    @pre_load
    def override_input(self, data):
        return self._override_input(data)

    def _override_input(self, data):
        pass

    @staticmethod
    def fix_badkey(data, badkey, goodkey):
        if badkey in data:
            warnings.warn(
                "{b} variable has been deprecated. Will try to fill in "
                "{g} if empty.".format(b=badkey, g=goodkey),
                DeprecationWarning)
            data[goodkey] = (
                data[goodkey] if (goodkey in data) else data[badkey])


class ProcessPoolParameters(argschema.schemas.DefaultSchema):
    pool_size = argschema.fields.Int(required=False, default=1)


class ZValueParameters(  # argschema.schemas.DefaultSchema,
                       OverridableParameterSchema):
    """template schema which interprets z values on which to act
    assumes a hierarchy such that minZ, maxZ are
    superceded by z which is superceded by zValues.
    """
    minZ = argschema.fields.Int(required=False)
    maxZ = argschema.fields.Int(required=False)
    z = argschema.fields.Int(required=False)
    zValues = argschema.fields.List(
            argschema.fields.Int,
            cli_as_single_argument=True,
            required=False)

    @post_load
    def generate_zValues(self, data):
        return self._generate_zValues(data)

    def _generate_zValues(self, data):
        if 'zValues' in data:
            return
        elif 'z' in data:
            data['zValues'] = [data['z']]
        elif ('minZ' in data) and ('maxZ' in data):
            data['zValues'] = range(data['minZ'], data['maxZ'] + 1)
        else:
            raise ValidationError("no z values specified")

    def _override_input(self, data):
        super(ZValueParameters, self)._override_input(data)
        # DEPRECATED
        # the following overrides should be removed in future versions
        self.fix_badkey(data, 'z_index', 'z')
        self.fix_badkey(data, 'zstart', 'minZ')
        self.fix_badkey(data, 'zend', 'maxZ')
        self.fix_badkey(data, 'zs', 'zValues')


class OutputStackParameters(RenderParameters, ZValueParameters,
                            ProcessPoolParameters, OverridableParameterSchema):
    """template schema for writing tilespecs to an output stack"""
    output_stack = argschema.fields.Str(required=True)
    close_stack = argschema.fields.Boolean(required=False, default=False)
    overwrite_zlayer = argschema.fields.Boolean(required=False, default=False)
    output_stackVersion = argschema.fields.Nested(
        RenderStackVersion, required=False)

    def _override_input(self, data):
        super(OutputStackParameters, self)._override_input(data)
        # DEPRECATED
        # the following overrides should be removed in future versions
        self.fix_badkey(data, 'stack', 'output_stack')
        self.fix_badkey(data, 'outputStack', 'output_stack')



from argschema import ArgSchemaParser
from .schemas import LensCorrectionSchema
from ..utils.generate_EM_tilespecs_from_metafile import \
        GenerateEMTileSpecsModule
from ..utils import utils as common_utils
from .mesh_and_solve_transform import MeshAndSolveTransform
from . import utils
from EMaligner import jsongz
import logging
import os
import glob
import json
import numpy as np
import renderapi
import cv2
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

logger = logging.getLogger(__name__)

example = {
        "data_dir": "/data/em-131fs3/lctest/T4_07/20190315142205_reference/0",
        "output_dir": "/data/em-131fs3/lctest/T4_07/20190315142205_reference/0",
        "mask_file": None,
        "ransac_thresh": 10,
        "nvertex": 1000,
        "regularization": {
            "default_lambda": 1.0,
            "lens_lambda": 1.0,
            "translation_factor": 1e-3
            },
        "log_level": "INFO",
        "write_pdf": False
        }


class LensCorrectionException(Exception):
    pass


def one_file(fdir, fstub):
    fullstub = os.path.join(fdir, fstub)
    files = glob.glob(fullstub)
    lf = len(files)
    if lf != 1:
        raise LensCorrectionException(
            "expected 1 file, found %d for %s" % (
                lf,
                fullstub))
    return files[0]


def tilespec_input_from_metafile(
        metafile, mask_file, output_dir, log_level, compress):
    result = {}
    result['metafile'] = metafile

    with open(metafile, 'r') as f:
        jmeta = json.load(f)

    result['z'] = np.random.randint(0, 1000)
    result['minimum_intensity'] = 0
    bpp = jmeta[0]['metadata']['camera_info']['camera_bpp']
    result['maximum_intensity'] = int(np.power(2, bpp) - 1)
    result['sectionId'] = jmeta[0]['metadata']['grid']
    result['maskUrl'] = mask_file
    result['output_path'] = os.path.join(output_dir, 'raw_tilespecs.json')
    result['log_level'] = log_level
    result['compress_output'] = compress
    return result


def make_collection_json(
        template_file,
        output_dir,
        thresh,
        compress,
        ignore_match_indices=None):

    with open(template_file, 'r') as f:
        matches = json.load(f)

    counts = []
    for m in matches['collection']:
        counts.append({})
        ind = np.arange(len(m['matches']['p'][0]))
        counts[-1]['n_from_gpu'] = ind.size

        _, _, w, _ = common_utils.pointmatch_filter(
                m,
                n_clusters=None,
                n_cluster_pts=20,
                ransacReprojThreshold=40.0,
                model='Similarity')

        m['matches']['w'] = w.tolist()

        counts[-1]['n_after_filter'] = np.count_nonzero(w)

    m = matches['collection']

    if ignore_match_indices:
        m = [match for i, match in enumerate(matches['collection'])
             if i not in ignore_match_indices]
        logger.warning("you are ignoring some point matches")

    collection_file = os.path.join(output_dir, "collection.json")
    collection_file = jsongz.dump(m, collection_file, compress=compress)

    return collection_file, counts


class LensCorrectionSolver(ArgSchemaParser):
    default_schema = LensCorrectionSchema

    def __init__(self, *args, **kwargs):
        super(LensCorrectionSolver, self).__init__(*args, **kwargs)
        self.jtform = None

    def run(self):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(self.args['log_level'])
        utils.logger.setLevel(self.args['log_level'])
        self.check_for_files()
        self.output_dir = self.args.get('output_dir', self.args['data_dir'])
        self.logger.info("destination directory:\n  %s" % self.output_dir)

        tspecin = tilespec_input_from_metafile(
                self.metafile,
                self.args['mask_file'],
                self.output_dir,
                self.args['log_level'],
                self.args['compress_output'])
        gentspecs = GenerateEMTileSpecsModule(input_data=tspecin, args=[])
        gentspecs.run()

        assert os.path.isfile(gentspecs.args['output_path'])
        self.logger.info(
                "raw tilespecs written:\n  %s" % gentspecs.args['output_path'])

        collection_path, self.filter_counts = make_collection_json(
                self.matchfile,
                self.output_dir,
                self.args['ransac_thresh'],
                self.args['compress_output'],
                self.args['ignore_match_indices'])

        self.n_from_gpu = np.array(
                [i['n_from_gpu'] for i in self.filter_counts]).sum()
        self.n_after_filter = np.array(
                [i['n_after_filter'] for i in self.filter_counts]).sum()
        self.logger.info(
                "filter counts: %0.2f %% kept" %
                (100 * float(self.n_after_filter) / self.n_from_gpu))

        assert os.path.isfile(collection_path)
        self.logger.info(
                "filtered collection written:\n  %s" % collection_path)

        solver_args = {
                'nvertex': self.args['nvertex'],
                'regularization': self.args['regularization'],
                'good_solve': self.args['good_solve'],
                'tilespec_file': gentspecs.args['output_path'],
                'match_file': collection_path,
                'output_dir': self.output_dir,
                'outfile': 'lens_correction_transform.json',
                'compress_output': self.args['compress_output'],
                'log_level': self.args['log_level']}

        self.solver = MeshAndSolveTransform(input_data=solver_args, args=[])
        self.solver.run()

        with open(self.solver.args['output_json'], 'r') as f:
            j = json.load(f)
        resolved_path = j['resolved_tiles']

        resolvedtiles = renderapi.resolvedtiles.ResolvedTiles(
                json=jsongz.load(resolved_path))

        self.jtform = resolvedtiles.transforms[0].to_dict()

        self.map1, self.map2, self.mask = utils.maps_from_tform(
                renderapi.transform.ThinPlateSplineTransform(
                    json=self.jtform),
                resolvedtiles.tilespecs[0].width,
                resolvedtiles.tilespecs[0].height,
                res=32)

        maskname = os.path.join(self.output_dir, 'mask.png')
        cv2.imwrite(maskname, self.mask)
        self.logger.info("wrote:\n  %s" % maskname)

        res = {}
        res['input'] = {}
        res['input']['template'] = os.path.abspath(self.matchfile)
        res['input']['metafile'] = os.path.abspath(self.metafile)
        res['output'] = {}
        res['output']['resolved_tiles'] = j.pop('resolved_tiles')
        res['output']['mask'] = os.path.abspath(maskname)
        res['output']['collection'] = os.path.abspath(collection_path)
        res['output']['lens_json'] = os.path.abspath(self.solver.tfpath)
        res['residual stats'] = j

        self.args['output_json'] = self.solver.args['output_json']

        with open(self.args['output_json'], 'w') as f:
            json.dump(res, f, indent=2)

    def check_for_files(self):
        self.metafile = one_file(self.args['data_dir'], '_metadata*')
        self.matchfile = one_file(self.args['data_dir'], '_template_matches*')
        self.logger.info("Using files: \n  %s\n  %s" % (
            self.metafile, self.matchfile))


if __name__ == '__main__':
    lcmod = LensCorrectionSolver(input_data=example)
    lcmod.run()
